package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet;

import java.util.Collection;
import java.util.Map;
import java.util.Set;

/** An abstract implementation of {@link Lattice} providing obvious
 * implementations of order-related operators and an 
 * implementation of {@link #elements()} based on exhaustive generation.
 * 
 * <p>The additional methods {@link #ensureElementsInLattice(Element[])}
 * and {@link #valueOfZeroOrOne(String)} can be used to simplify concrete
 * implementing subclasses.
 */

public abstract class AbstractLattice implements Lattice {

	public boolean isDistributive() {
		return false;
	}

	/**
	 * Ensure that the provided element belong to this lattice, throwing an
	 * {@link ElementLatticeMismatchException} otherwise.
	 * 
	 * @param x an element.
	 */

	protected void ensureElementsInLattice( Element x ) {
		if ( x.lattice() != this ) throw new ElementLatticeMismatchException( x, this );
	}

	/**
	 * Ensure that the provided elements belong to this lattice, throwing an
	 * {@link ElementLatticeMismatchException} otherwise.
	 * 
	 * @param x an element.
	 * @param y another element.
	 */

	protected void ensureElementsInLattice( Element x, Element y ) {
		if ( x.lattice() != this ) throw new ElementLatticeMismatchException( x, this );
		if ( y.lattice() != this ) throw new ElementLatticeMismatchException( y, this );
	}

	/**
	 * Ensure that the provided elements belong to this lattice, throwing an
	 * {@link ElementLatticeMismatchException} otherwise.
	 * 
	 * @param element some elements.
	 */

	protected void ensureElementsInLattice( Element... element ) {
		for ( Element e : element )
			if ( e.lattice() != this ) throw new ElementLatticeMismatchException( e, this );
	}

	/**
	 * A simple support method to make it easy to parse zero and one.
	 * 
	 * @param s a string.
	 * @return {@link Lattice#zero()} or {@link Lattice#one()} is <code>s</code> is <code>0</code>
	 * or <code>1</code>, respectively; <code>null</code> otherwise.
	 */

	protected Element valueOfZeroOrOne( final String s ) {
		final String t = s.trim();
		if ( "0".equals( t ) ) return zero();
		if ( "1".equals( t ) ) return one();
		return null;
	}

	/** Generate iteratively all elements of this lattice.
	 * 
	 * <p>This methods uses {@link Lattice#generators()} to obtain an initial set of elements,
	 * and then computes joins and meets of all available elements until no new elements
	 * are generated.
	 * It is expected that concrete subclasses will override this method 
	 * with an <i>ad hoc</i>, more efficient implementation.
	 * 
	 * <p>It is strongly suggested that concrete subclasses that do not override this method
	 * cache its result internally.
	 * 
	 * @return the set of elements of this lattice. 
	 */
	
	public Collection<Element> elements() {
		final Set<Element> elements = new ObjectOpenHashSet<Element>( generators() );
		Set<Element> generatedElements = new ObjectOpenHashSet<Element>( elements );
		Set<Element> newElements;
		
		int size;
		Element zero = zero(), one = one();
		elements.add( zero );
		elements.add( one );

		do {
			newElements = generatedElements;
			generatedElements = new ObjectOpenHashSet<Element>();

			size = elements.size();
			// We need to extract elements into an array, as it will be modified by the loop
			for ( Element x: elements.toArray( new Element[ elements.size() ] ) ) {
				if ( x == zero || x == one ) continue;
				for ( Element y: newElements ) {
					if ( x == y || y == zero || y == one ) continue;
					Element join = x.join( y );
					// We can skip comparable elements
					if ( join.equals( x ) || join.equals( y ) ) continue;
					Element meet = x.meet( y );
					if ( elements.add( join ) ) generatedElements.add( join );
					if ( elements.add( meet ) ) generatedElements.add( meet );
				}
			}
		} while ( size != elements.size() );

		return elements;
	}

	/** Check whether two elements are comparable using {@link Lattice#meet(Element[])}.
	 * 
	 * @param x an element.
	 * @param y another element.
	 * @return true iff <code>x.meet(y)</code> equals <code>x</code> or <code>y</code>.
	 */
	
	public boolean comp( Element x, Element y ) {
		ensureElementsInLattice( x, y );
		final Element meet = x.meet( y );
		return meet.equals( x ) || meet.equals( y );
	}

	/** Check whether an element is smaller than or equal to another element using {@link Lattice#meet(Element[])}.
	 * 
	 * @param x an element.
	 * @param y another element.
	 * @return true iff <code>x.meet(y)</code> equals <code>x</code>.
	 */
	public boolean leq( Element x, Element y ) {
		ensureElementsInLattice( x, y );
		return x.meet( y ).equals( x );
	}
	
	/** Enumerate all elements of this lattice to compute explicitly the
	 * difference using the definition given in {@link Lattice#psdiff(Element, Element)}.
	 * It is expected that concrete subclasses will override this method 
	 * with an <i>ad hoc</i>, more efficient implementation.
	 * 
	 * @see Lattice#psdiff(Element, Element)
	 */
	
	public Element psdiff( final Element x, final Element y ) {
		if ( ! isDistributive() ) throw new UnsupportedOperationException();
		Element curr = one();
		for( Element e: elements() )
			if ( x.leq( y.join( e ) ) ) curr = curr.meet( e );
		return curr;
	}

	/** Enumerate all elements of this lattice to compute explicitly the
	 * pseudocomplement using the definition given in {@link Lattice#pscomp(Element, Element)}.
	 * It is expected that concrete subclasses will override this method 
	 * with an <i>ad hoc</i>, more efficient implementation.
	 * 
	 * @see Lattice#pscomp(Element, Element)
	 */
	
	public Element pscomp( final Element x, final Element y ) {	
		if ( ! isDistributive() ) throw new UnsupportedOperationException();
		Element curr = zero();
		for( Element e: elements() )
			if ( x.meet( e ).leq( y ) ) curr = curr.join( e );
		return curr;
	}

	/** {@inheritDoc} */
	
	public Element symdiff( final Element x, final Element y ) {
		if ( ! isDistributive() ) throw new UnsupportedOperationException();
		return psdiff( x, y ).join( psdiff( y, x ) );
	}


	/** Return the covering relation of this lattice.
	 *
	 * <p>This method examines exhaustively all pairs of elements <var>x</var>, <var>y</var>
	 * such that there is no element strictly between <var>x</var> and <var>y</var>. In the
	 * worst case, this requires cubic time.
	 * It is expected that concrete subclasses will override this method 
	 * with an <i>ad hoc</i>, more efficient implementation.
	 * 
	 * @return a map from elements to set of elements representing the covering relation.
	 */
	
	public Map<Element,Set<Element>> coveringRelation() {
		Collection<Element> elements = elements();
		Object2ObjectOpenHashMap<Element,Set<Element>> map = new Object2ObjectOpenHashMap<Element,Set<Element>>(); 
		for( Element x : elements )
			for( Element y : elements ) {
				// Test only if x < y (strict!)
				if ( x == y || ! x.leq( y ) ) continue;
				
				boolean middle = false;
				for( Element z : elements ) {
					if ( x != z && y != z && x.leq( z ) && z.leq( y ) ) {
						middle = true;
						break;
					}
				}
				
				if ( ! middle ) {
					Set<Element> curr = map.get( x );
					if ( curr == null ) {
						curr = new ObjectOpenHashSet<Element>();
						curr.add( y );
						map.put( x, curr );
					}
					else curr.add( y );
				}
			}
		return map;
	}
}
