package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.IntOpenHashSet;
import it.unimi.dsi.fastutil.ints.IntSet;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;

import java.util.Collection;

/**
 * The distributive lattice of subsets of an initial segment of the natural numbers.
 */
public class Parts extends AbstractDistributiveLattice {
	private final int n;

	private final Subset zero;

	private final Subset one;

	private final static class Subset extends AbstractElement {
		final private IntOpenHashSet elements;

		public Subset( Lattice lattice, int... element ) {
			super( lattice );
			this.elements = new IntOpenHashSet( element );
		}

		public Subset( Lattice lattice, IntSet elements ) {
			super( lattice );
			this.elements = new IntOpenHashSet( elements );
		}

		public String toString() {
			return elements.toString();
		}

		public boolean equals( Object o ) {
			if ( o instanceof Subset ) {
				Subset subset = (Subset)o;
				return subset.lattice == lattice && subset.elements.equals( elements );
			}
			else return false;
		}

		public int hashCode() {
			return lattice.hashCode() ^ elements.hashCode();
		}
	}


	public Parts( int n ) {
		this.n = n;
		zero = new Subset( this );
		int[] all = new int[ n ];
		for ( int i = n; i-- != 0; )
			all[ i ] = i;
		one = new Subset( this, all );
	}

	public Element valueOf( String name ) {
		throw new IllegalArgumentException();
	}

	/** Return an element of this lattice corresponding
	 * to the specified subset.
	 * 
	 * @param element a list of integers interpreted as a subset.
	 * @return an element of this lattice corresponding
	 * to the specified subset.
	 */
	
	public Element valueOf( int... element ) {
		for ( int e : element )
			if ( e < 0 || e >= n ) throw new IllegalArgumentException( Integer.toString( e ) );
		return new Subset( this, element );
	}

	public Collection<Element> generators() {
		ObjectArrayList<Element> generators = ObjectArrayList.wrap( new Element[ n ], 0 );
		for ( int i = 0; i < n; i++ ) generators.add( new Subset( this, i ) );
		return generators;
	}

	public Element one() {
		return one;
	}

	public Element zero() {
		return zero;
	}

	public Element join( Element... elements ) {
		IntSet result = new IntOpenHashSet();
		for ( Element e : elements )
			result.addAll( ( (Subset)e ).elements );
		return new Subset( this, result );
	}

	public Element meet( Element... elements ) {
		IntSet result = new IntOpenHashSet( one.elements );
		for ( Element e : elements )
			result.retainAll( ( (Subset)e ).elements );
		return new Subset( this, result );
	}

	/** Compute the pseudo-difference of two elements.
	 * 
	 * <p>In the case of a Boolean algebra, <var>x</var> &minus; <var>y</var> = <var>x</var> &#x2227; ~<var>y</var>.
	 * @see Lattice#psdiff(Element, Element)
	 */
	
	@Override
	public Element psdiff( final Element x, final Element y ) {
		IntOpenHashSet result = new IntOpenHashSet( ((Subset)x).elements );
		result.removeAll( ((Subset)y).elements );
		return new Subset( this, result.toIntArray() );
	}
	
	/** Compute the set implication of two elements.
	 * 
	 * <p>In the case of a Boolean algebra, <var>x</var> &#8658; <var>y</var> = ~<var>x</var> &#x2228; <var>y</var>.
	 */

	@Override
	public Element pscomp( final Element x, final Element y ) {
		IntOpenHashSet result = new IntOpenHashSet( ((Subset)y).elements );
		final IntOpenHashSet set = ((Subset)x).elements;
		for( int i = n; i-- != 0; ) if ( ! set.contains( i ) ) result.add( i );
		return new Subset( this, result.toIntArray() );
	}

	@Override
	public String toString() {
		return "Parts(" + n + ")";
	}
}
