package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet;

import java.util.Arrays;
import java.util.Collection;
import java.util.Set;

/**
 * The (categorical) product of lattices. As for every algebraic theory, elements are
 * tuples, and operations are implemented componentwise.
 * 
 * <p>Note that the product of zero lattices is the trivial lattice with one element.
 */
public class Product extends AbstractLattice {

	/** Cached set of elements. */
	private Set<Element> elements;
	/** The component lattices. */
	private final Lattice[] lattice;
	/** True off all elements in {@link #lattice} are distributive. */
	private final boolean isDistributive;
	
	private final Tuple zero;
	private final Tuple one;

	private final static class Tuple extends AbstractElement {

		public final Element[] component;

		public Tuple( Product lattice, Element... element ) {
			super( lattice );
			for ( int i = 0; i < element.length; i++ )
				if ( ! lattice.component( i ).equals( element[ i ].lattice() ) ) throw new IllegalArgumentException();
			component = new Element[ element.length ];
			System.arraycopy( element, 0, component, 0, element.length );
		}

		@Override
		public String toString() {
			StringBuilder ris = new StringBuilder( "<" );
			for ( Element x : component ) {
				// debug
				if ( x.toString() == null ) System.err.println( "e che cazz" );
				// debgu end
				ris.append( x.toString() + "," );
			}
			ris.deleteCharAt( ris.lastIndexOf( "," ) );
			ris.append( ">" );
			return ris.toString();
		}

		@Override
		public boolean equals( Object o ) {
			if ( o instanceof Tuple ) {
				Tuple v = (Tuple)o;
				if ( !v.lattice().equals( lattice ) ) return false;
				return Arrays.equals( v.component, component );
			}
			else return false;
		}

		@Override
		public int hashCode() {
			return lattice.hashCode() ^ Arrays.hashCode( component );
		}
	}

	/** Return the product of the provided lattices.
	 * 
	 * @param lattice the lattices whose product has to be computed.
	 * @return the product of the specified lattices.
	 */
	public static Lattice of( Lattice... lattice ) {
		if ( lattice.length == 0 ) return new Trivial();
		return new Product( lattice );
	}

	private Product( Lattice... lattice ) {
		if ( lattice.length == 0 ) throw new IllegalArgumentException();
		this.lattice = new Lattice[ lattice.length ];

		System.arraycopy( lattice, 0, this.lattice, 0, lattice.length );

		boolean t = true;
		for ( int i = lattice.length; i-- != 0; ) t &= lattice[ i ].isDistributive();
		isDistributive = t;

		Element[] e = new Element[ lattice.length ];
		for ( int i = e.length; i-- != 0; )
			e[ i ] = lattice[ i ].zero();
		zero = new Tuple( this, e );
		for ( int i = e.length; i-- != 0; )
			e[ i ] = lattice[ i ].one();
		one = new Tuple( this, e );
	}

	@Override
	public boolean isDistributive() {
		return isDistributive;
	}
	
	@Override
	public Collection<Element> elements() {
		if ( elements == null ) {
			elements = new ObjectOpenHashSet<Element>();
			generateElements( 0, new Element[ lattice.length ], lattice, elements );
		}
		
		return elements;
	}

	public Collection<Element> generators() {
		final Set<Element> generators = new ObjectOpenHashSet<Element>();
		generateGenerators( 0, new Element[ lattice.length ], lattice, generators );
		generators.remove( zero );
		generators.remove( one );
		return generators;
	}

	/** Project the provided element on a component specified by index.
	 * 
	 * @param e an element of this product.
	 * @param i a component, specified by its index starting at 0.
	 * @return the projection of <code>e</code> on the <code>i</code>-th component.
	 */
	
	public Element proj( Element e, int i ) {
		ensureElementsInLattice( e );
		return ((Tuple)e).component[ i ];
	}
	
	public Element valueOf( String name ) {
		throw new ElementNameException( name, this );
	}

	/** Return an element of this product given a tuple of elements
	 * from its factors.
	 * 
	 * @param element a list of elements from each factor of this product, in the same order.
	 * @return the element of the product defined by the tuple <code>element</code>.
	 */
	
	public Element tuple( Element... element ) {
		return new Tuple( this, element );
	}

	private void generateElements( int n, Element[] curr, Lattice[] lattice, Set<Element> tuples ) {
		if ( n == lattice.length - 1 ) {
			for ( Element x : lattice[ n ].elements() ) {
				curr[ n ] = x;
				tuples.add( new Tuple( this, curr ) );
			}
			return;
		}

		for ( Element x : lattice[ n ].elements() ) {
			curr[ n ] = x;
			generateElements( n + 1, curr, lattice, tuples );
		}
	}

	private void generateGenerators( int n, Element[] curr, Lattice[] lattice, Set<Element> tuples ) {
		if ( n == lattice.length - 1 ) {
			for ( Element x : lattice[ n ].generators() ) {
				curr[ n ] = x;
				tuples.add( new Tuple( this, curr ) );
			}

			// This is necessary to get an actual set of generators.
			curr[ n ] = lattice[ n ].zero();
			tuples.add( new Tuple( this, curr ) );
			return;
		}

		for ( Element x : lattice[ n ].elements() ) {
			curr[ n ] = x;
			generateGenerators( n + 1, curr, lattice, tuples );
		}

		// As above.
		curr[ n ] = lattice[ n ].zero();
		generateGenerators( n + 1, curr, lattice, tuples );
	}

	public Element one() {
		return one;
	}

	public Element zero() {
		return zero;
	}

	public Element join( Element... element ) {
		ensureElementsInLattice( element );

		if ( element.length == 0 ) return zero();
		else if ( element.length == 1 ) return element[ 0 ];
		else {
			Element[] ris = new Element[ ( (Tuple)element[ 0 ] ).component.length ];
			for ( int i = 0; i < ris.length; i++ ) {
				Element[] tmp = new Element[ element.length ];
				for ( int j = 0; j < element.length; j++ )
					tmp[ j ] = ( (Tuple)element[ j ] ).component[ i ];
				ris[ i ] = lattice[ i ].join( tmp );
			}
			return new Tuple( this, ris );
		}
	}

	public Element meet( Element... element ) {
		ensureElementsInLattice( element );

		if ( element.length == 0 ) return one();
		else if ( element.length == 1 ) return element[ 0 ];
		else {
			Element[] ris = new Element[ ( (Tuple)element[ 0 ] ).component.length ];
			for ( int i = 0; i < ris.length; i++ ) {
				Element[] tmp = new Element[ element.length ];
				for ( int j = 0; j < element.length; j++ )
					tmp[ j ] = ( (Tuple)element[ j ] ).component[ i ];
				ris[ i ] = lattice[ i ].meet( tmp );
			}
			return new Tuple( this, ris );
		}
	}

	/**
	 * Return a component lattice, given its index.
	 * 
	 * <p>The component lattices of a product are numbered from 0. This method provides access by
	 * index to such components.
	 * 
	 * @param k a component index.
	 * @return the component lattice of index <code>k</code>.
	 */

	public Lattice component( final int k ) {
		return lattice[ k ];
	}

	public boolean comp( Element x, Element y ) {
		ensureElementsInLattice( x, y );
		boolean leq = true, geq = true;
		Tuple t = (Tuple)x, u = (Tuple)y;
		for ( int i = t.component.length; i-- != 0; ) {
			leq &= t.component[ i ].leq( u.component[ i ] );
			geq &= u.component[ i ].leq( t.component[ i ] );
		}
		return leq || geq;
	}

	public boolean leq( Element x, Element y ) {
		ensureElementsInLattice( x, y );
		boolean result = true;
		Tuple t = (Tuple)x, u = (Tuple)y;
		for ( int i = t.component.length; i-- != 0; )
			result &= t.component[ i ].leq( u.component[ i ] );
		return result;
	}

	public String toString() {
		StringBuilder s = new StringBuilder();
		s.append( "( " );
		for( int i = 0; i < lattice.length; i++ ) {
			if ( i != 0 ) s.append( Lattice.UTF8 ? " \u00D7 " : " * " );
			s.append( lattice[ i ] );
		}
				
		return s.append( " )" ).toString();
	}
}
