package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.objects.ObjectArrayList;
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet;
import it.unimi.dsi.fastutil.objects.ObjectSets;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.Set;

/** The (categorical) sum (a.k.a&#46; <em>free direct product</em>) of distributive lattices.
 * 
 * <p>
 * As for every algebraic theory, elements are terms built from the elements of
 * the two lattices, and quotiented w.r.t. all equations of the theory and all
 * equations that are true in either summand. Each element of a summand is thus
 * identified with an element of the sum, which can be obtained 
 * by {@linkplain #inj(int,Element) injection}.
 * 
 * <p>
 * In this implementation, to avoid equivalence classes and make the sum computable
 * we restrict to distributive lattices and use Gr&#228;tzer&ndash;Lakser's
 * (disjunctive) canonical representatives (see &ldquo;Chain conditions in the 
 * distributive free product of lattices&rdquo;, <i>Trans. Amer. Math. Soc.</i>, 
 * 144:301&minus;312, 1969) described in clausal form. 
 * 
 * <P>
 * A <em>clause</em> is a set of elements
 * from the summands. A <em>reduced clause</em> is a clause containing at most one element
 * from each summand. A (clausal) <em>formula</em> is a set of clauses.
 * 
 * <p>
 * Formulae can be interpreted in a disjunctive or conjunctive way 
 * (i.e., as a join of meets or a meet of joins),
 * and both intepretations are used during computations. A (disjunctive or conjunctive) 
 * <em>reduced formula</em> is a set of reduced clauses in which no clause can be eliminated
 * without affecting the formula's value. Elements of a sum
 * are stored as disjunctive reduced formulae.
 * 
 * <p><strong>Warning</strong>: if one of the summands is trivial (just one element)
 * it will <strong>not</strong> annihilate the sum (as it should happen) but
 * rather it will behave like a neutral element (a r&ocirc;le actually
 * played by the two-element Boolean lattice). This unfortunate restriction
 * is caused by the difficulty of detecting whether one of the summands is trivial (as
 * {@link Lattice#elements()} is optional).  
 */
public class Sum extends AbstractDistributiveLattice {
	/** The number of summands of this sum. */
	private final int n;
	/** The summands of this sum. */
	private final Lattice[] summand;
	/** An array parallel to {@link #summand} containing the name of each lattice (for {@link #toString()}). */
	private final String[] name;
	/** Cached set of elements. */
	private Collection<Element> elements;

	private final DisjunctiveFormula zero;
	private final DisjunctiveFormula one;
	
	private static enum ClauseType {
		DISJUNCTIVE,
		CONJUNCTIVE
	}
	
	/** Return the sum of the given distributive lattices.
	 * 
	 * @param distributiveLattice the distributive lattices whose sum has to be computed.
	 * @return the sum of the specified distributive lattices.
	 */
	
	public static Lattice of( final Lattice... distributiveLattice ) {
		if ( distributiveLattice.length > 0 ) return new Sum( distributiveLattice, null );
		return new Boolean();
	}

	/** Return the sum of the given distributive lattices, giving them specified names.
	 * 
	 * @param distributiveLattice the distributive lattices whose sum has to be computed.
	 * @param name a parallel list of names for each <code>distributiveLattice</code>.
	 * @return the sum of the specified distributive lattices.
	 */
	
	public static Lattice of( final Lattice[] distributiveLattice, final String[] name ) {
		if ( distributiveLattice.length > 0 ) return new Sum( distributiveLattice, name );
		return new Boolean();
	}

	@SuppressWarnings("unchecked")
	private Sum( final Lattice[] distributiveLattice, final String[] name ) {
		if ( distributiveLattice.length == 0 ) throw new IllegalArgumentException( Integer.toString( distributiveLattice.length ) );
		if ( name != null && distributiveLattice.length != name.length ) throw new IllegalArgumentException( "You have provided " + name.length + " names for " + distributiveLattice.length + " lattices" );
		
		for ( Lattice l : distributiveLattice )
			if ( ! l.isDistributive() ) throw new IllegalArgumentException( "Non-distributive summands are not allowed: " + l );

		this.summand = distributiveLattice;
		this.n = summand.length;
		this.name = new String[ summand.length ];
		for( int i = summand.length; i-- != 0; ) this.name[ i ] = name != null ? name[ i ] : Integer.toString( i );
		
		zero = new DisjunctiveFormula( this, Collections.EMPTY_SET );
		one = new DisjunctiveFormula(this, ObjectSets.singleton( new Clause( new Element[ n ] ) ) );
	}

	public Collection<Element> generators() {
		ObjectArrayList<Element> result = new ObjectArrayList<Element>();
		for( int i = n; i-- != 0; )
			for ( Element e : summand[ i ].generators() )
				result.add( new DisjunctiveFormula( this, i, e ) );
		return result;
	}

	@Override
	public Collection<Element> elements() {
		if ( elements == null ) elements = super.elements();
		return elements;
	}
	
	public Element valueOf( String name ) {
		Element zeroOrOne = valueOfZeroOrOne( name );
		if ( zeroOrOne != null ) return zeroOrOne;
		throw new ElementNameException( name, this );
	}

	/** Inject an element of a summand into this sum.
	 * 
	 * @param k the index of a summand.
	 * @param element an element of the <code>k</code>-th summand.
	 * @return the corresponding element of this sum.
	 */
	public Element inj( int k, Element element ) {
		if ( summand[ k ] != element.lattice() ) throw new IllegalArgumentException( "Element " + element + " belongs to lattice " + element.lattice() + ", but the summand of index " + k + " is " + summand[ k ] );
		// Summands zeroes and ones must be injected into our zero and one.
		if ( element.equals( element.lattice().zero() ) ) return zero;
		if ( element.equals( element.lattice().one() ) ) return one;
		return new DisjunctiveFormula( this, k, element );
	}
	
	public Element zero() {
		return zero;
	}

	public Element one() {
		return one;
	}

	private final static class Clause {
		/** The elements of this clause. Entries are indexed in parallel with {@link Sum#summand},
		 * and missing elements are <code>null</code>. */
		public final Element[] element;
		/** Cached hash code. */
		private int hashCode;
	
		public Clause( final Element[] element ) {
			this.element = element;
		}
		
		public int hashCode() {
			if ( hashCode == 0 ) hashCode = Arrays.hashCode( element );
			return hashCode;
		}
		
		public boolean equals( Object o ) {
			if ( o instanceof Clause ) {
				final Clause clause = (Clause)o;
				if ( clause.hashCode() != hashCode() ) return false;
				return Arrays.equals( element, clause.element );
			}
			return false;
		}

		public boolean isEmpty() {
			for( int i = element.length; i-- != 0; ) if ( element[ i ] != null ) return false;
			return true;
		}

		public int size() {
			int c = 0;
			for( int i = element.length; i-- != 0; ) if ( element[ i ] != null ) c++;
			return c;
		}
}
	
	private final static class DisjunctiveFormula extends AbstractElement {
		/** The set of conjunctive clauses making up this formula. */
		public final Set<Clause> clauses;
		/** Cached hash code. */
		private int hashCode;
		
		protected DisjunctiveFormula( Sum sum, Set<Clause> clauses ) {
			super( sum );
			this.clauses = clauses;
		}

		protected DisjunctiveFormula( Sum sum, int k, Element element ) {
			super( sum );
			if ( sum.summand[ k ] != element.lattice() ) throw new ElementLatticeMismatchException( element, sum.summand[ k ] );
			Element[] e = new Element[ sum.n ];
			e[ k ] = element;
			clauses = ObjectSets.singleton( new Clause( e ) );
		}

		public String toString() {
			if ( clauses.isEmpty() ) return "0";
			
			StringBuilder s = new StringBuilder();
			boolean firstClause = true;
			for( Clause clause : clauses ) {
				if ( ! firstClause ) {
					if ( Lattice.UTF8 ) s.append( " \u2228 " );
					else if ( Lattice.RING ) s.append( " + " );
					else s.append( " | " );
				}
				
				if ( clause.isEmpty() ) s.append( 1 );
				else {
					boolean firstElement = true;
					if ( clause.size() > 1 && clauses.size() > 1 ) s.append( "( " );
					for( int i = 0; i < clause.element.length; i++ ) {
						if ( clause.element[ i ] != null ) {
							if ( ! firstElement ) {
								if ( Lattice.UTF8 ) s.append( " \u2227 " );
								else if ( Lattice.RING ) s.append( " * " );
								else s.append( " & " );
							}
							String t = clause.element[ i ].toString();
							// Let us try to make sums of sums readable
							if ( ((Sum)lattice).summand[ i ] instanceof Sum ) t =  "[ " + t + " ]"; 
							s.append( t + "_" + ((Sum)lattice).name[ i ] );
							firstElement = false;
						}
					}
					if ( clause.size() > 1 && clauses.size() > 1 ) s.append( " )" );
				}
				
				firstClause = false;
			}
			return s.toString();
		}

		public boolean equals( Object o ) {
			if ( o instanceof DisjunctiveFormula ) {
				DisjunctiveFormula formula = (DisjunctiveFormula)o;
				if ( formula.hashCode() != hashCode() ) return false;
				return formula.lattice().equals( lattice ) && formula.clauses.equals( clauses );
			}
				
			return false;
		}

		public int hashCode() {
			if ( hashCode == 0 ) hashCode = lattice.hashCode() ^ clauses.hashCode();
			return hashCode;
		}
	}

	/** Compute a disjunctive representation of a meet of disjunctive formulae by distributing
	 * the meet over the internal joins and gathering the resulting meets of meets by associativity.
	 * The gathering process includes reduction.
	 * 
	 * @param formulae a set of disjunctive formulae, to be intepreted as a meet.
	 * @return a disjunctive formula equivalent to <code>formulae</code>. 
	 */
	@SuppressWarnings("unchecked")
	private Set<Clause> distributeAndGather( Set<Set<Clause>> formulae ) {
		final Set<Clause> result = new ObjectOpenHashSet<Clause>();
		distributeAndGatherRec( formulae.toArray( new Set[ formulae.size() ] ), 0, new Clause[ formulae.size() ], result );
		return result;
	}
	
	private void distributeAndGatherRec( Set<Clause>[] set, int s, Clause[] current, Set<Clause> result ) {
		if ( s == set.length ) {
			/* We enumerate the chosen clauses, reduce them to a single clause, and check them for maximality
			  against the current clauses in result. */
			Element[] meet = new Element[ n ];
			for( Clause clause : current )
				for( int i = n; i-- != 0; )
					if ( clause.element[ i ] != null ) 
						meet[ i ] = meet[ i ] == null ? clause.element[ i ] : ( meet[ i ].meet( clause.element[ i ] ) );
			
			final Clause c = new Clause( meet );
			boolean ok = true;
			for( Iterator<Clause> i = result.iterator(); i.hasNext(); ) {
				Clause m = i.next();
				if ( conjClauseLeq( c, m ) ) {
					ok = false;
					break;
				}
				if ( conjClauseLeq( m, c ) ) i.remove();
			}						
			
			if ( ok ) result.add( c );
			return;
		}

		for ( Clause e : set[ s ] ) {
			current[ s ] = e;
			distributeAndGatherRec( set, s + 1, current, result );
		}
	}
	

	/** Applie a choice function (i.e., distributes) a formula and then applies reduction to the
	 * resulting formula.
	 * 
	 * @param formula a formula.
	 * @param clauseType how the clauses resulting from choice should be interpreted; note
	 * that this is dual to the interpretation of clauses in <code>formula</code>.
	 * @return a formula obtained by taking as new clauses sets of elements obtained by picking in
	 * all possible ways an element from each clause in <code>formula</code>. The resulting clauses are considered
	 * conjunctive or disjunctive, depending on <code>clauseType</code>, and reduced accordingly.
	 */
	private Set<Clause> chooseAndReduce( final Set<Clause> formula, final ClauseType clauseType ) {
		final Set<Clause> result = new ObjectOpenHashSet<Clause>();
		chooseAndReduceRec( formula.toArray( new Clause[ formula.size() ] ), 0, new Element[ formula.size() ], new int[ formula.size() ], result, clauseType );
		return result;
	}
	
	private <T> void chooseAndReduceRec( final Clause[] clause, final int s, final Element[] current, final int[] index, final Set<Clause> result, final ClauseType clauseType ) {
		if ( s == clause.length ) {
			final Element[] element = new Element[ n ];
			for( int i = current.length; i-- != 0; ) {
				final Element e = current[ i ];
				final int k = index[ i ];
				element[ k ] = element[ k ] == null ? e : ( clauseType == ClauseType.DISJUNCTIVE ? e.join( element[ k ] ) : e.meet( element[ k ] ) );
			}

			
			Clause c = new Clause( element );
			
			// We now scan the known clauses, deleting redundant ones.

			for( Iterator<Clause> i = result.iterator(); i.hasNext(); ) {
				Clause d = i.next();
				if ( clauseType == ClauseType.DISJUNCTIVE ) {
					if ( disjClauseLeq( d, c ) ) {
						c = null;
						break;
					}
					if ( disjClauseLeq( c, d ) ) i.remove();
				}
				else {
					if ( conjClauseLeq( c, d ) ) {
						c = null;
						break;
					}
					if ( conjClauseLeq( d, c ) ) i.remove();
				}
			}
			
			if ( c != null ) result.add( c );
			return;
		}

		for ( int i = clause[ s ].element.length; i-- != 0; ) {
			Element e = clause[ s ].element[ i ];
			if ( e == null ) continue;
			current[ s ] = e;
			index[ s ] = i;
			chooseAndReduceRec( clause, s + 1, current, index, result, clauseType );
		}
	}
	
	
	/** Check whether two disjunctive clauses are in order. For this to
	 * happen, and for each key of the first map the corresponding
	 * element of the second map must be present and larger or equal.
	 * 
	 * @param x a disjunctive clause.
	 * @param y a disjunctive clause.
	 * @return whether <code>x</code> &le; <code>y</code>.
	 */
	private boolean disjClauseLeq( final Clause x, final Clause y ) {
		Element[] a = x.element, b = y.element;
		Element e, f;
		for( int i = n; i-- != 0; ) {
			e = a[ i ];
			f = b[ i ];
			if ( e == null ) continue;
			if ( e != null && f == null || ! e.leq ( f ) ) return false;
		}
		
		return true;
	}

	
	/** Check whether two conjunctive clauses are in order. For this to
	 * happen, for each key of the second map the corresponding
	 * element of the first map must present and smaller or equal.
	 * 
	 * @param x a conjunctive clause.
	 * @param y a conjunctive clause.
	 * @return whether <code>x</code> &le; <code>y</code>.
	 */

	private boolean conjClauseLeq( final Clause x, final Clause y ) {
		Element[] a = x.element, b = y.element;
		Element e, f;
		for( int i = n; i-- != 0; ) {
			e = a[ i ];
			f = b[ i ];
			if ( f == null ) continue;
			if ( e == null && f != null || ! e.leq ( f ) ) return false;
		}
		
		return true;
	}


	public Element meet( Element... element ) {
		ensureElementsInLattice( element );
		// First we build a conjunctive set of formulae.
		Set<Set<Clause>> formulae = new ObjectOpenHashSet<Set<Clause>>();
		for ( Element e : element ) formulae.add( ( (DisjunctiveFormula)e ).clauses );
		// Then, we distribute and gather, and canonicalise.
		return new DisjunctiveFormula( this, chooseAndReduce( chooseAndReduce( distributeAndGather( formulae ), ClauseType.DISJUNCTIVE ), ClauseType.CONJUNCTIVE ) );
	}

	public Element join( Element... element ) {
		ensureElementsInLattice( element );
		// First we build a large formula by taking the union of all elements (discarding nonmaximal clauses).
		Set<Clause> result = new ObjectOpenHashSet<Clause>();
		for ( Element e : element ) {
			for( Clause clause : ( (DisjunctiveFormula)e ).clauses ) {
				boolean ok = true;
				for( Iterator<Clause> i = result.iterator(); i.hasNext(); ) {
					Clause m = i.next();
					if ( conjClauseLeq( clause, m ) ) {
						ok = false;
						break;
					}
					if ( conjClauseLeq( m, clause ) ) i.remove();
				}
				if ( ok ) result.add( clause );	
			}
		}
		// Then, we canonicalise.
		return new DisjunctiveFormula( this, chooseAndReduce( chooseAndReduce( result, ClauseType.DISJUNCTIVE ), ClauseType.CONJUNCTIVE ) );
	}
	
	public String toString() {
		StringBuilder s = new StringBuilder();
		s.append( "( " );
		for( int i = 0; i < summand.length; i++ ) {
			if ( i != 0 ) s.append( " + " );
			s.append( summand[ i ] );
		}
				
		return s.append( " )" ).toString();
	}
	
	public boolean comp( Element x, Element y ) {
		return x.leq( y ) || y.leq( x );
	}

	public boolean leq( Element x, Element y ) {
		ensureElementsInLattice( x, y );
		if ( y.equals( one ) || x.equals( zero ) ) return true;
		// x <= y iff for each clause in x there is a corresponding larger clause in y (see paper).
		Set<Clause> a = ((DisjunctiveFormula)x).clauses, b = ((DisjunctiveFormula)y).clauses; 
		
		for( Clause ma : a ) {
			boolean foundClause = false;
			for( Clause mb : b ) 
				if ( conjClauseLeq( ma, mb ) ) {
					foundClause = true;
					break;
				}
			if ( foundClause == false ) return false;
		}
		
		return true;
	}
}
