package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import static org.junit.Assert.assertEquals;

import java.util.Collection;

import org.junit.Test;

public class ProductTest extends LaMa4JTestCase {

	/** Assert that a specified projection of a product ends in a specified lattice
	 * and is a lattice morphism.
	 * 
	 * @param p a product.
	 * @param i an index.
	 * @param l a lattice.
	 */
	public static void assertProjection( Product p, int i, Lattice l ) {
		Collection<Element> elements = p.elements();

		assertEquals( l.zero(), p.proj( p.zero(), i ) );
		assertEquals( l.one(), p.proj( p.one(), i ) );

		for( Element x : elements )
			for( Element y : elements ) {
				assertEquals( p.proj( x.meet( y ), i ), p.proj( x, i ).meet( p.proj( y, i ) ) );
				assertEquals( p.proj( x.join( y ), i ), p.proj( x, i ).join( p.proj( y, i ) ) );
			}
	}

	
	@Test
	public void testProductOfChains() {
		Chain l = new Chain(3, "x" );
		Chain m = new Chain(4, "a", "b" );

		Product p = (Product)Product.of( l, m );

		assertLattice( p );
		assertProjection( p, 0, l );
		assertProjection( p, 1, m );
		
		assertEquals( 12, p.elements().size() );
	}

	@Test
	public void testProductOfParts() {
		Parts l = new Parts( 2 );
		Parts m = new Parts( 3 );

		Product p = (Product)Product.of( l, m );

		assertLattice( p );
		assertProjection( p, 0, l );
		assertProjection( p, 1, m );

		assertEquals( 32, p.elements().size() );
	}

	@Test
	public void testNested() {
		Boolean b = new Boolean();

		Product a = (Product)Product.of( b, b );
		assertProjection( a, 0, b );
		assertProjection( a, 1, b );
		
		Product p = (Product)Product.of( a, a );
		assertProjection( p, 0, a );
		assertProjection( p, 1, a );

		assertLattice( p );

		assertEquals( 16, p.elements().size() );
	}

	@Test
	public void testUnit() {
		Chain l = new Chain(3, "x" );
		Trivial t = new Trivial();
		
		Product p = (Product)Product.of( l, t );

		assertLattice( p );
		assertProjection( p, 0, l );
		assertProjection( p, 1, t );
		
		assertEquals( 3, p.elements().size() );
	}
}
