package it.unimi.dsi.lama4j;

/*		 
 * LaMa4J: Lattice Manipulation for Java
 *
 * Copyright (C) 2006-2016 Marco Rosa and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import static org.junit.Assert.assertEquals;

import java.util.Collection;

import org.junit.Test;

public class SumTest extends LaMa4JTestCase {
	
	/** Assert that a specified lattice injects in a specified sum, and that 
	 * the injection is a lattice morphism.
	 * 
	 * @param l a lattice.
	 * @param s a sum.
	 */
	public static void assertInjection( Lattice l, int k, Sum s ) {
		Collection<Element> elements = l.elements();

		assertEquals( s.inj( k, l.one() ), s.one() );
		assertEquals( s.inj( k, l.zero() ), s.zero() );

		for( Element x : elements )
			for( Element y : elements ) {
				assertEquals( s.inj( k, x.meet( y ) ), s.inj( k, x ).meet( s.inj( k, y ) ) );
				assertEquals( s.inj( k, x.join( y ) ), s.inj( k, x ).join( s.inj( k, y ) ) );
			}
	}
	
	
	@Test
	public void testSum() {
		Chain l = new Chain(3, "x" );
		Chain m = new Chain(4, "a", "b" );

		Sum s = (Sum)Sum.of( l, m );

		assertLattice( s );
		assertInjection( l, 0, s );
		assertInjection( m, 1, s );
		
		Element x = s.inj( 0, l.valueOf( 1 ) );
		Element a = s.inj( 1, m.valueOf( 1 ) );
		Element b = s.inj( 1, m.valueOf( 2 ) );
		
		assertEquals( x, x.join( a ).meet( x ) );
		assertEquals( x, x.meet( a ).join( x ) );
		assertEquals( x.meet( b ).join( a ), x.join( a ).meet( b ) );
		
	}
	

	@Test
	public void testSumEqual() {
		Chain l = new Chain(3, "x" );

		Sum s = (Sum)Sum.of( l, l );

		assertLattice( s );
		assertInjection( l, 0, s );
		assertInjection( l, 1, s );
		
		Element x = s.inj( 0, l.valueOf( 1 ) );
		Element a = s.inj( 1, l.valueOf( 1 ) );
		
		assertEquals( x, x.join( a ).meet( x ) );
		assertEquals( x, x.meet( a ).join( x ) );
	}
	

	@Test
	public void testNested() {
		Chain l = new Chain(3, "x" );
		Chain m = new Chain(3, "y" );
		
		Sum s = (Sum)Sum.of( l, m );
		assertInjection( l, 0, s );
		assertInjection( m, 1, s );

		Sum t = (Sum)Sum.of( s, l );
		assertInjection( s, 0, t );
		assertInjection( l, 1, t );

		assertLattice( t );

		assertEquals( 20, t.elements().size() );
	}

	@Test
	public void testUnit() {
		Chain l = new Chain(3, "x" );
		Boolean b = new Boolean();

		Sum s = (Sum)Sum.of( l, b );
		
		assertLattice( s );
		assertInjection( l, 0, s );
		assertInjection( b, 1, s );

		assertEquals( l.elements().size(), s.elements().size() );
	}

}
